/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/** sign extentions                            **/
/************************************************/
DATA_TYPE sext(DATA_TYPE d){
  return (d & BIT31) ? d | EXTND32 : d & MASK32;
}
/************************************************/
DATA_TYPE sext16(DATA_TYPE d){
  return (d & BIT15) ? d | EXTND16 : d & MASK16;
}
/************************************************/
DATA_TYPE sext21(DATA_TYPE d){
  return (d & BIT20) ? d | EXTND21 : d & MASK21;
}

/************************************************/
uint_64t bzap(uint_64t Rav, int byte_mask){
  if(byte_mask & BIT0) Rav &= ZAP0;
  if(byte_mask & BIT1) Rav &= ZAP1;
  if(byte_mask & BIT2) Rav &= ZAP2;
  if(byte_mask & BIT3) Rav &= ZAP3;
  if(byte_mask & BIT4) Rav &= ZAP4;
  if(byte_mask & BIT5) Rav &= ZAP5;
  if(byte_mask & BIT6) Rav &= ZAP6;
  if(byte_mask & BIT7) Rav &= ZAP7;
  return Rav;
}

/** 0x16  IEEE floating-point instruction      **/
/************************************************/
void FuFLTI(INST_TYPE ir, DATA_TYPE Rav, 
	      DATA_TYPE Rbv, DATA_TYPE *w){
  int CI= get_code(ir);
  DATA_TYPE Rcv=0;
  int func4  = (ir>>5) & 0xF;
  int func11 = (ir>>5) & 0x7FF;
  double Fav=0.0, Fbv=0.0, Fcv=0.0;
  memcpy(&Fav, &Rav, 8);
  memcpy(&Fbv, &Rbv, 8);

  switch(func4){
  case 0x0: Fcv=Fav + Fbv; break;
  case 0x1: Fcv=Fav - Fbv; break;
  case 0x2: Fcv=Fav * Fbv; break;
  case 0x5: Fcv=(Fav==Fbv) ? 2.0 : 0.0; break;
  case 0x6: Fcv=(Fav< Fbv) ? 2.0 : 0.0; break;
  case 0x7: Fcv=(Fav<=Fbv) ? 2.0 : 0.0; break;
  case 0xe: Fcv=(double)(int_64t)Rbv; break;
  case 0x4: 
    Fcv=(isnan(Fav) || isnan(Fbv)) ? 2.0 : 0.0; 
    break;
  case 0xc: { // CVTTS
    memcpy(&Fcv, &Rbv, 8);
    float tmp = (float)Fcv;
    Fcv = tmp;
    break;
  }
  case 0x3: // DIVT
    if(Fbv==0.0){ Fcv = 0.0; break; }
    Fcv = (double)(Fav / Fbv);
    break;
  case 0xf: { // CVTTQ
    int_64t tmp;
    tmp = (int_64t)Fbv;
    memcpy(&Fcv, &tmp, 8);
    break;
  }
  default: 
    printf("** SimAlpha FuFLTI %x\n", CI);
    exit(1);
  }
  
  switch(func11){
  case 0x0bc:
  case 0x03c:
  case 0x07c:
  case 0x0fc: 
  case 0x7bc: 
  case 0x73c: 
  case 0x77c: 
  case 0x7fc: // CVTQS
    Fcv = (double)(int_64t)Rbv;
    break;
  }

  memcpy(&Rcv, &Fcv, 8);
  *w = Rcv;
}

/** 0x1c  Floating-point to integer register   **/
/************************************************/
void FuFPTI(INST_TYPE ir, DATA_TYPE Rav, 
	      DATA_TYPE Rbv, DATA_TYPE *w){
  int CI= get_code(ir);
  DATA_TYPE Rcv=0;
  int func11 = (ir>>5) & 0x7FF;
  double Fav=0.0, Fbv=0.0;
  memcpy(&Fav, &Rav, 8);
  memcpy(&Fbv, &Rbv, 8);

  switch(func11){
  case 0x000: 
    Rcv = (Rbv & BIT7) ? 
      Rbv |= EXTND8 : (Rbv & 0xffull);
    break;
  case 0x001: // SEXTW(Sign Extend Word)
    Rcv = (Rbv & BIT15) ? 
      Rbv |= EXTND16 : (Rbv & 0xffffull);
    break;
  case 0x070: // FTOIT: Rc <= Fav
    Rcv = Rav;
    break;
  case 0x078: // FTOIS: Rc <= Fav
    Rcv = ((Rav>>32) & 0xc0000000) 
      | ((Rav >> 29) & 0x3fffffff);
    if(Rav & BIT63) Rcv |= EXTND32;
    break;
    
  default: 
    printf("** SimAlpha FuFPTI %x\n", CI);
    exit(1);
  }
  *w = Rcv;
}

/** 0x17  Floating-point instruction           **/
/************************************************/
void FuFLTL(INST_TYPE ir, DATA_TYPE Rav, 
	      DATA_TYPE Rbv, DATA_TYPE *w){
  int CI= get_code(ir);
  DATA_TYPE Rcv=0;
  int func11 = (ir>>5) & 0x7FF;
  double Fav=0.0, Fbv=0.0, Fcv=0.0;
  memcpy(&Fav, &Rav, 8);
  memcpy(&Fbv, &Rbv, 8);

  switch(func11){
  case 0x020: { /*__CPYS;  Copy Sign */
    uint_64t tmp=0;
    tmp |= (Rav &  BIT63);
    tmp |= (Rbv & ~BIT63);
    memcpy(&Fcv, &tmp, 8);
    break;
  }
  case 0x021: { /*__CPYSN; Copy Sign Negative */
    uint_64t tmp=0;
    tmp |= (~Rav &  BIT63);
    tmp |= ( Rbv & ~BIT63);
    memcpy(&Fcv, &tmp, 8);
    break;
  }
    /*__CPYSE; Copy Sign and Exponent */
  case 0x022: {
    uint_64t tmp=0;
    tmp |= (Rav &  0xFFF0000000000000ull);
    tmp |= (Rbv & ~0xFFF0000000000000ull);
    memcpy(&Fcv, &tmp, 8);
    break;
  }
  case 0x010: { /*__CVTLQ; */
    uint_64t tmp = 0;
    tmp  = 0;
    tmp |= ((Rbv>>32) & 0xc0000000ull);
    tmp |= ((Rbv>>29) & 0x3fffffffull);
    if(tmp & BIT31) tmp |= EXTND32;
    memcpy(&Fcv, &tmp, 8);
    break;
  }
  case 0x030: { /*__CVTQL; */
    DATA_TYPE tmp = 0;
    // SimpleScalar has a bug?
    tmp |=  (((Rbv << 32) & 
	      0xc000000000000000ull));
    tmp |=  (((Rbv << 29) & 
	      0x07ffffffe0000000ull));
    memcpy(&Fcv, &tmp, 8);
    break;
  }
  case 0x24: /* __MT_FPCR */
    /* Move Fav to Floating-point 
       Control Register
       Return 0.0 for the simplicity.
    */
    Fcv=0.0; break;
  case 0x25: /* __MF_FPCR */
    /* Move from Floating-point 
       Control Register to Fa
       Return 0.0 for the simplicity.
    */
    Fcv=0.0; break;
  default:
    printf("** SimAlpha FuFLTL %x\n", CI);
    exit(1);
  }
  memcpy(&Rcv, &Fcv, 8);
  *w = Rcv;
}

/** 0x14  Integer to floating-point register   **/
/************************************************/
void FuITFP(INST_TYPE ir, DATA_TYPE Rav, 
	      DATA_TYPE Rbv, DATA_TYPE *w){
  int CI= get_code(ir);
  DATA_TYPE Rcv=0;
  int func11 = (ir>>5) & 0x7FF;
  double Fav=0.0, Fbv=0.0, Fcv=0.0;
  memcpy(&Fav, &Rav, 8);
  memcpy(&Fbv, &Rbv, 8);

  switch(func11){
  case 0x024: // ITOFT
    memcpy(&Fcv, &Rav, 8);
    break;
    
  case 0x004: { // ITOFS
    DATA_TYPE tmp = 0;
    tmp = ((Rav & BIT31) << 32) | 
      (map_s(Rav) << 52) |
      ((Rav & 0x7fffff) << 29);
    memcpy(&Fcv, &tmp, 8);
    break;
  }
  case 0x0ab: // SQRTT
    Fcv = sqrt(Fbv);
    break;
    
  default: 
    printf("** SimAlpha FuITFP %x\n", CI);
    exit(1);
  }
  memcpy(&Rcv, &Fcv, 8);
  *w = Rcv;
}

/** Integer Logic Unit                         **/
/************************************************/
void UnitL(int CI, DATA_TYPE a, 
	   DATA_TYPE b, DATA_TYPE *c){
  switch(CI){
  case _BIS___: *c = a |  b; break;
  case _AND___: *c = a &  b; break;
  case _BIC___: *c = a & ~b; break;
  case _ORNOT_: *c = a | ~b; break;
  case _XOR___: *c = a ^  b; break;
  case _EQV___: *c = a ^ ~b; break;
  case _AMASK_: *c=b & ~CPU_FEATURE_MASK; break; 
  case _IMPLVER:*c=IMPLEMENTATION_VERSION; break;
  default:
    printf("** SimAlpha AluLogi %x\n", CI);
  }
}

/** Integer Arithmetic                         **/
/************************************************/
void UnitA(int CI, DATA_TYPE a, 
	   DATA_TYPE b, DATA_TYPE *c){
  switch(CI){
  case _S4ADDQ: *c = (a<<2) + b;       break;
  case _ADDQ__: *c = a      + b;       break;
  case _SUBQ__: *c = a      - b;       break;
  case _ADDL__: *c = sext(a      + b); break;
  case _S4ADDL: *c = sext((a<<2) + b); break;
  case _SUBL__: *c = sext(a      - b); break;
  case _S4SUBL: *c = sext((a<<2) - b); break;
  case _S8ADDL: *c = sext((a<<3) + b); break;
  case _S8SUBL: *c = sext((a<<3) - b); break;
  case _CMPULT: *c = (a < b);          break;
  case _S4SUBQ: *c = (a<<2) - b;       break;
  case _S8ADDQ: *c = (a<<3) + b;       break;
  case _S8SUBQ: *c = (a<<3) - b;       break;
  case _CMPEQ_: *c=((int_64t)a==(int_64t)b);break;
  case _CMPULE: *c=(a <= b);                break;
  case _CMPLT_: *c=((int_64t)a< (int_64t)b);break;
  case _CMPLE_: *c=((int_64t)a<=(int_64t)b);break;
  case _CMPBGE:
    *c=0;
    if((a & ~ZAP0)>=(b & ~ZAP0)) *c|=BIT0;
    if((a & ~ZAP1)>=(b & ~ZAP1)) *c|=BIT1;
    if((a & ~ZAP2)>=(b & ~ZAP2)) *c|=BIT2;
    if((a & ~ZAP3)>=(b & ~ZAP3)) *c|=BIT3;
    if((a & ~ZAP4)>=(b & ~ZAP4)) *c|=BIT4;
    if((a & ~ZAP5)>=(b & ~ZAP5)) *c|=BIT5;
    if((a & ~ZAP6)>=(b & ~ZAP6)) *c|=BIT6;
    if((a & ~ZAP7)>=(b & ~ZAP7)) *c|=BIT7;
    break;
    
  default:
    printf("** SimAlpha AluUnit %x\n", CI);
  }
}

/************************************************/
inline DATA_TYPE B4(DATA_TYPE b){
  return  ((b&0x7)==0) ? 0 : 64 - ((b&0x7)<<3);
}

inline DATA_TYPE B3(DATA_TYPE b){ 
  return (b & 0x7);
}

inline DATA_TYPE B7(DATA_TYPE b){ 
  return (b & 0x7)<<3;
}

/************************************************/
void UnitS(int CI, DATA_TYPE a, 
	   DATA_TYPE b, DATA_TYPE *c){
  uint_32t msk, loc; /* byte_mask, byte_loc */
  switch(CI){
  case _SLL___: *c = a<<(b & 0x3F); break;
  case _SRL___: *c = a>>(b & 0x3F); break;
  case _SRA___: *c =(int_64t)a>>(b & 0x3F); break;
  case _ZAPNOT: *c = bzap(a, (int)~b); break;
  case _ZAP___: *c = bzap(a, (int)b ); break;
  case _EXTBL_: *c = (a>>B7(b)) & MASK08; break;
  case _EXTWL_: *c = (a>>B7(b)) & MASK16; break;
  case _EXTLL_: *c = (a>>B7(b)) & MASK32; break;
  case _EXTQL_: *c =  a>>B7(b); break;
  case _EXTWH_: *c = (a<<B4(b)) & MASK16; break;
  case _EXTLH_: *c = (a<<B4(b)) & MASK32; break;
  case _EXTQH_: *c =  a<<B4(b); break;
  case _MSKBL_: *c=bzap(a, 0x01u<<B3(b)); break;
  case _MSKWL_: *c=bzap(a, 0x03u<<B3(b)); break;
  case _MSKLL_: *c=bzap(a, 0x0fu<<B3(b)); break;
  case _MSKQL_: *c=bzap(a, 0xffu<<B3(b)) ;break;
  case _MSKWH_: *c = bzap(a, (0x03u<<B3(b))>>8); 
    break;
  case _MSKLH_: *c = bzap(a, (0x0fu<<B3(b))>>8); 
    break;
  case _MSKQH_: *c = bzap(a, (0xffu<<B3(b))>>8); 
    break;
  case _INSBL_: 
    *c =bzap(a<<B7(b), ~(0x01u<<B3(b))); 
    break;
  case _INSWL_: 
    *c =bzap(a<<B7(b), ~(0x03u<<B3(b))); 
    break;
  case _INSLL_: 
    *c =bzap(a<<B7(b), ~(0x0fu<<B3(b))); 
    break;
  case _INSQL_: 
    *c =bzap(a<<B7(b), ~(0xffu<<B3(b))); 
    break;
  case _INSWH_:
    msk = (uint_32t)((0x3ull << (b & 0x7))>>8);
    loc = (uint_32t)(64 - ((b & 0x7)<<3)) & 0x3f;
    *c = bzap((a >> loc), ~msk);
    break;
  case _INSLH_:
    msk = (uint_32t)((0xFull << (b & 0x7))>>8);
    loc = (uint_32t)(64 - ((b & 0x7)<<3)) & 0x3f;
    *c = bzap((a >> loc), ~msk);
    break;
  case _INSQH_:
    msk = (uint_32t)((0xFFull << (b & 0x7))>>8);
    loc = (uint_32t)(64 - ((b & 0x7)<<3)) & 0x3f;
    *c = bzap((a >> loc), ~msk);
    break;
  default:
    printf("** SimAlpha ShiftUnit %x\n", CI);
  }
}

/** Convert double to uint_64t                 **/
/************************************************/
double Fcv(DATA_TYPE a){
  double Fav;
  memcpy(&Fav, &a, 8);
  return Fav;
}

/** 64 bit multiply Input of A and B           **/
/**  output of High 64b and Low 64b            **/
/************************************************/
void mul128(DATA_TYPE a, DATA_TYPE b, 
	    DATA_TYPE *rh, DATA_TYPE *rl){
  DATA_TYPE ah,al, bh, bl;
  DATA_TYPE x,y,z;
  int carry=0;   /* has a value of 0, 1 or 2 */
  
  ah = a >> 32;
  al = a & 0xFFFFFFFF;
  
  bh = b >> 32;
  bl = b & 0xFFFFFFFF;
  
  x = al*bl;
  y = (al*bh << 32);
  z = x + y;
  if(z<x || z<y) carry ++;
  x = z;
  y = (ah*bl << 32);
  z = x + y;
  if(z<x || z<y) carry ++;
  *rl = z;
  
  *rh = ah*bh+(al*bh >> 32)+(ah*bl >> 32)+carry;
}

/************************************************/
void UnitM(int CI, DATA_TYPE a, 
	   DATA_TYPE b, DATA_TYPE *c){
  DATA_TYPE tmp2, tmp3;
  switch(CI){
  case _MULL__: *c = sext(a*b);         break;
  case _MULQ__: mul128(a, b, &tmp2, c); break;
  case _UMULH_: mul128(a, b, c, &tmp3); break;
  default:
    *c = 0;
    printf("** SimAlpha MulUnit %x\n", CI);
  }
}

/** Integer Arithmetic and Logic Unit          **/
/************************************************/
void CmovUnit(int CI, DATA_TYPE a, int *cm){
  double Fav =0.0;
  switch(CI){
  case _CMOVEQ : *cm=(a==0);          break;
  case _CMOVNE : *cm=(a!=0);          break;
  case _CMOVLT : *cm=((int_64t)a< 0); break;
  case _CMOVGE : *cm=((int_64t)a>=0); break;
  case _CMOVLE : *cm=((int_64t)a<=0); break;
  case _CMOVGT : *cm=((int_64t)a> 0); break;
  case _CMOVLBC: *cm=(!(a & 1));      break;
  case _CMOVLBS: *cm=(int)(a & 1);    break;
  case _FCMOVEQ:
    memcpy(&Fav, &a, 8);
    *cm = (Fav==0.0);
    break;
  case _FCMOVNE:
    memcpy(&Fav, &a, 8);
    *cm = (Fav!=0.0);
    break;
  default:
    printf("** SimAlpha CmovUnit %x\n", CI);
  }
}

/** calculate branch condition : taken?        **/
/************************************************/
int Btaken(int CI, DATA_TYPE a){
  int taken = 0;
  switch(CI){
  case _BEQ_: taken = (a==0);          break;
  case _BNE_: taken = (a!=0);          break;
  case _BR__: taken = 1;               break;
  case _BSR_: taken = 1;               break;
  case _BLE_: taken = ((int_64t)a<=0); break;
  case _BLT_: taken = ((int_64t)a< 0); break;
  case _BGE_: taken = ((int_64t)a>=0); break;
  case _BGT_: taken = ((int_64t)a> 0); break;
  case _BLGC: taken = (!(a & 1));      break;
  case _BLBS: taken = (int)(a & 1 );   break;
  case _FBEQ: taken=(Fcv(a)==0.0);     break;
  case _FBLT: taken=(Fcv(a)< 0.0);     break;
  case _FBLE: taken=(Fcv(a)<=0.0);     break;
  case _FBNE: taken=(Fcv(a)!=0.0);     break;
  case _FBGE: taken=(Fcv(a)>=0.0);     break;
  case _FBGT: taken=(Fcv(a)> 0.0);     break;
  default: printf("** SimAlpha Btaken %x\n", CI);
  }
  return taken;
}
/************************************************/
