/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/************************************************/
simple_chip::simple_chip(char *prog, char **opt){
  ev      = new Env();
  ev->sc  = new system_config(prog, opt);
  ev->e   = new evaluation_result();
  ev->as  = new architecture_state(ev);
  ev->mem = new memory_system(ev);
  ev->deb = new debug(ev);
  ev->sys = new system_manager(ev);
  p       = new instruction(ev);
}

/************************************************/
simple_chip::~simple_chip(){   /** destructor **/
  print_evaluation_result(ev);
}

/************************************************/
int simple_chip::step(){
  p->Fetch(&ev->as->pc);   /* pipeline stage 0 */
  p->Slot();               /* pipeline stage 1 */
  p->Issue();              /* pipeline stage 3 */
  p->RegisterRead();       /* pipeline stage 4 */
  p->Execute();            /* pipeline stage 5 */
  p->Memory();             /* pipeline stage 6 */
  p->WriteBack();
  
  ev->e->retired_inst++;
  house_keeper(p);
  
  return ev->sys->running;
}

#define IB_MSK 0x0ffff      /* mask of inst_buf */
/** main loop of the optimized version         **/
/************************************************/
void simple_chip::loop(){
  instruction **ib = new instruction*[IB_MSK+1];
  for(int i=0; i<IB_MSK+1; i++){
    ib[i] = new instruction(ev);
  }
  
  while(ev->sys->running){
    int index = (ev->as->pc>>2) & IB_MSK;
    instruction *pt = ib[index];
    
    if(pt->Cpc!=ev->as->pc){
      pt->Fetch(&ev->as->pc);
      pt->Slot();
      pt->Issue();
    }
    pt->BackEnd();

    ev->e->retired_inst++;
    if(ev->sc->slow_mode) house_keeper(pt);
   }
}
/************************************************/
