/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"
#define R_ZERO 31   /** Index of zero register **/

/************************************************/
int INTx_OP(int Op){/* 0x10-0x14 OP_INTx */
  return ((Op&MSK2)==0x10);
}

/* return 1 if data is written to integer reg.  */
/************************************************/
int WR(int Op){ 
  return !((Op&MSK2)==0x14 || (Op&MSK2)==0x20);
}

/************************************************/
int LDAx(int CI){
  return (CI==_LDA___ || CI==_LDAH__);
}

/************************************************/
instruction::instruction(Env *ev_t){
  ev = ev_t;
}

/************************************************/
void instruction::Fetch(data_t *pc){
  ev->mem->ld_inst(pc, &ir);
  Cpc = *pc;
  Npc = *pc + 4;
}

/** decode the instruction                     **/
/************************************************/
void instruction::Slot(){ 
  CI = get_code(ir);      /* Code ID           */
  Op = (ir>>26) & 0x3f;   /* Op field          */
  CM = cmov_ir(ir);       /* set if CMOV       */
  BR = ((Op&MSK4)==0x30); /* 0x30-0x3f branch  */
  LD = ld_bytes(CI);      /* not 0 if load     */
  ST = st_bytes(CI);      /* not 0 if store    */

  int Rc = ir & 0x1f;     /* Rc field          */
  int Ra = (ir>>21)&0x1f; /* Ra field          */
  int Rb = (ir>>16)&0x1f; /* Rb field          */
  int Misc = (Op==OP_MISC || Op==OP_PAL);
  /***************** Rav ********************/
  Ai = LDAx(CI);
  RA = (Ai || LD || Misc) ? R_ZERO : Ra;
  Ar = !Ai & 
    !(Op==OP_FLTV || Op==OP_FLTI ||
      Op==OP_FLTL || Op==OP_FPTI ||
      (Op&MSK2)==0x24 || /* 0x24-27 STfloat  */
      (Op&MSK3)==0x30);  /* 0x30-37 BR & FBR */

  /***************** Rbv ********************/
  Bi = BR || (INTx_OP(Op) && (ir & BIT12));
  RB = (Bi || Misc) ? R_ZERO : Rb;
  Br = !Bi && !((Op&MSK2)==0x14); /* 0x14-17 */

  /***************** WriteBack index ********/
  WB = (LD || LDAx(CI) || Op==OP_JSR || 
	CI==_BR__ || CI==_BSR_) ? Ra :
    ((Op&MSK3)==0x10 || /* 0x10-0x17 */
     Op==OP_FPTI) ? Rc : R_ZERO;
}

/************************************************/
void instruction::Rename(){ }

/** genarate the immediate data Imm            **/
/************************************************/
void instruction::Issue(){ 
  DATA_TYPE Lit = (ir >> 13) & MASK08;
  DATA_TYPE D21 = sext21(ir) << 2;
  DATA_TYPE D16 = (Op!=OP_LDAH) ? sext16(ir) :
    sext16(ir) << 16;
  
  Imm = (LD || ST || LDAx(CI)) ? D16 :
    (BR) ? D21 : (Bi) ? Lit : 0;
}

/** copy data to Rav and Rbv                   **/
/************************************************/
void instruction::RegisterRead(){
  Rav = Ar ? ev->as->r[RA] : Ai ? Imm : 
    ev->as->f[RA];
  Rbv = Br ? ev->as->r[RB] : Bi ? Imm : 
    ev->as->f[RB];
  if(CM){
    Adr = WR(Op) ? ev->as->r[WB] : ev->as->f[WB];
  }
}

/** update Rcv, Adr and Npc                    **/
/************************************************/
void instruction::Execute(){
  if(LD || ST){
    Adr = (CI==_LDQ_U_ || CI==_STQ_U_) ?
      (Rbv + Imm) & ~7 : Rbv + Imm;
  }
  else if(BR){
    Rcv = Cpc+4;
    Npc = Btaken(CI, Rav) ? Cpc+4+Rbv : Cpc+4;
  }
  else if(CM){ /** Conditional move inst. **/
    int cm=0;
    CmovUnit(CI, Rav, &cm);
    Rcv = (cm) ? Rbv : Adr;
  }
  else{
    switch(Op){
    case OP_JSR : 
      Npc = Rbv & ~3ull; 
      Rcv = Cpc + 4; 
      break;
    case OP_LDA : Rcv = Rav + Rbv;          break;
    case OP_LDAH: Rcv = Rav + Rbv;          break;
    case OP_INTA: UnitA(CI, Rav, Rbv, &Rcv);break;
    case OP_INTL: UnitL(CI, Rav, Rbv, &Rcv);break;
    case OP_INTS: UnitS(CI, Rav, Rbv, &Rcv);break;
    case OP_INTM: UnitM(CI, Rav, Rbv, &Rcv);break;
    case OP_ITFP: FuITFP(ir,Rav, Rbv, &Rcv);break;
    case OP_FLTL: FuFLTL(ir,Rav, Rbv, &Rcv);break;
    case OP_FPTI: FuFPTI(ir,Rav, Rbv, &Rcv);break;
    case OP_FLTI: FuFLTI(ir,Rav, Rbv, &Rcv);break;
    case OP_PAL : Rcv=0;                    break;
    case OP_MISC: Rcv=0;                    break;
    default: printf("** ExeUnit_def: %x", CI);
    }
  }
  ev->as->pc = Npc;
}

/**  store Rav to mem or load to Rcv           **/
/************************************************/
void instruction::Memory(){ 
  if(LD){
    data_t tmp;
    ev->mem->mm->ld_8byte(&Adr, &tmp);
    set_ld_data(CI, LD, &Adr, &tmp, &Rcv);
  }
  else if(ST){
    data_t tmp;
    DATA_TYPE mask;
    set_st_data(CI, ST, &Adr, &Rav, &tmp, &mask);
    ev->mem->mm->st_8byte(&Adr, &tmp, mask);
  }
}

/** PAL or copy Rcv into the regfile           **/
/************************************************/
void instruction::WriteBack(){ 
  if(Op==OP_PAL) ev->sys->execute_pal(this);
  if(WB!=R_ZERO){
    if(WR(Op)) ev->as->r[WB] = Rcv;
    else       ev->as->f[WB] = Rcv;
  }
}

/** back-end stages for the optimized version  **/
/************************************************/
void instruction::BackEnd(){
  /***** RegisterRead  **************************/
  Rav = Ar ? ev->as->r[RA] : Ai ? Imm : 
    ev->as->f[RA];
  Rbv = Br ? ev->as->r[RB] : Bi ? Imm : 
    ev->as->f[RB];
  if(CM){
    Adr = WR(Op) ? ev->as->r[WB] : ev->as->f[WB];
  }

  /***** Execute ********************************/
  if(LD || ST){
    Adr = (CI==_LDQ_U_ || CI==_STQ_U_) ?
      (Imm+Rbv) & ~7 : Imm+Rbv;
  }
  else if(BR){
    Rcv = Cpc+4;
    Npc = Btaken(CI, Rav) ? Cpc+4+Rbv : Cpc+4;
  }
  else if(CM){ /** Conditional move inst. **/
    int cm=0;
    CmovUnit(CI, Rav, &cm);
    Rcv = (cm) ? Rbv : Adr;
  }
  else{
    switch(Op){
    case OP_JSR : 
      Npc = Rbv & ~3ull; 
      Rcv = Cpc + 4; 
      break;
    case OP_LDA : Rcv = Rav + Rbv;          break;
    case OP_LDAH: Rcv = Rav + Rbv;          break;
    case OP_INTA: UnitA(CI, Rav, Rbv, &Rcv);break;
    case OP_INTL: UnitL(CI, Rav, Rbv, &Rcv);break;
    case OP_INTS: UnitS(CI, Rav, Rbv, &Rcv);break;
    case OP_INTM: UnitM(CI, Rav, Rbv, &Rcv);break;
    case OP_ITFP: FuITFP(ir,Rav, Rbv, &Rcv);break;
    case OP_FLTL: FuFLTL(ir,Rav, Rbv, &Rcv);break;
    case OP_FPTI: FuFPTI(ir,Rav, Rbv, &Rcv);break;
    case OP_FLTI: FuFLTI(ir,Rav, Rbv, &Rcv);break;
    case OP_PAL : Rcv=0;                    break;
    case OP_MISC: Rcv=0;                    break;
    default: printf("** ExeUnit_def: %x", CI);
    }
  }
  ev->as->pc = Npc;

  /***** Memory *********************************/
  if(LD){
    data_t tmp;
    ev->mem->mm->ld_8byte(&Adr, &tmp);
    set_ld_data(CI, LD, &Adr, &tmp, &Rcv);
  }
  else if(ST){
    data_t tmp;
    DATA_TYPE mask;
    set_st_data(CI, ST, &Adr, &Rav, &tmp, &mask);
    ev->mem->mm->st_8byte(&Adr, &tmp, mask);
  }

  /***** WriteBack ******************************/
  if(Op==OP_PAL) ev->sys->execute_pal(this);
  if(WB!=R_ZERO){
    if(WR(Op)) ev->as->r[WB]=Rcv;
    else       ev->as->f[WB]=Rcv;
  }
}
/************************************************/
