/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/************************************************/
int ld_bytes(int CI){
  switch(CI){
  case _LDQ___: return 8;
  case _LDQ_L_: return 8;
  case _LDQ_U_: return 8;
  case _LDG___: return 8;
  case _LDT___: return 8;
  case _LDF___: return 8;
  case _LDL___: return 4;
  case _LDS___: return 4;
  case _LDL_L_: return 4;
  case _OPC0C_: return 2;
  case _OPC0A_: return 1;
  }
  return 0;
}

/************************************************/
int st_bytes(int CI){
  switch(CI){
  case _STQ___: return 8;
  case _STQ_C_: return 8;
  case _STQ_U_: return 8;
  case _STG___: return 8;
  case _STT___: return 8;
  case _STF___: return 8;
  case _STL___: return 4;
  case _STS___: return 4;
  case _STL_C_: return 4;
  case _OPC0D_: return 2;
  case _OPC0E_: return 1;
  }
  return 0;
}

/************************************************/
void set_ld_data(int CI, int n, data_t *adr, 
		 data_t *in, data_t *out){
  int off8 = (*adr & 7) << 3;
  switch(n){
  case 4:
    if(CI==_LDL___ || CI==_LDL_L_){
      *out = sext((*in >> off8) & MASK32);
    }
    else if(CI==_LDS___){
      DATA_TYPE tmp = (*in >> off8) & MASK32;
      DATA_TYPE rt  = tmp;
      rt = (tmp >> 31) & 0x1ull;
      rt = (rt << 11) | (map_s(tmp) & 0x7ffull);
      rt = (rt << 23) | (tmp & 0x7FFFFFull);
      rt =  rt << 29;  
      *out = rt;
    }
    break;
  case 8: *out = *in; break;
  case 2: *out = (*in >> off8) & MASK16; break;
  case 1: *out = (*in >> off8) & MASK08; break;
  }
}

/************************************************/
void set_st_data(int CI, int n, data_t *adr, 
		 data_t *in, data_t *out, 
		 DATA_TYPE *msk){
  int off8 = (*adr & 7) << 3;
  if(CI==_STS___){
    DATA_TYPE tmp;
    uint_32t intt = 0;
    intt |= ((*in >> 29) & 0x3FFFFFFFul);
    intt |= ((*in >> 32) & 0xc0000000ul);
    memcpy(&tmp, &intt, 4);
    *out = (tmp & MASK32) << off8;
    *msk = ~(MASK32 << off8);
  }
  else{
    switch(n){
    case 4:
      *out = (*in & MASK32) << off8;
      *msk = ~(MASK32 << off8);
      break;
    case 8: 
      *out = *in;
      *msk = 0; 
      break;
    case 2:
      *out = (*in & MASK16) << off8;
      *msk = ~(MASK16 << off8);
      break;
    case 1:
      *out = (*in & MASK08) << off8;
      *msk = ~(MASK08 << off8);
      break;
    }
  }
}

/************************************************/
inline uint_32t gen_tag(uint_32t adr){
  return (adr  & 0xffffffc0);
}

/************************************************/
void memory_system::ld_inst(data_t *a, 
			    INST_TYPE *ir){
  data_t d;
  mm->ld_8byte(a, &d);
  if(*a%8==0) *ir =  d      & MASK32;
  if(*a%8==4) *ir = (d>>32) & MASK32;
}

/************************************************/
void memory_system::ld_8byte(data_t *a, data_t *d){
  mm->ld_8byte(a, d);
}

void memory_system::st_8byte(data_t *a, data_t *d,
			     DATA_TYPE msk){
  mm->st_8byte(a, d, msk);
}

/************************************************/
memory_system::~memory_system(){}

/************************************************/
memory_system::memory_system(Env *ev_t){
  ev = ev_t;
  mm = new main_memory(ev);

  char *prog_name = ev->sc->program_name;

  /*** Read memory image from the file ***/
  FILE *fp;
  if((fp = fopen(prog_name, "r")) == NULL) {
    fprintf(stderr, "Bad file name: %s\n", 
	    prog_name);
    exit(1);
  }
  char buf[4096];
  fgets(buf, 4096, fp); /** File Format Check **/
  if(strncmp(buf+3, "SimAlpha", 8)){
    fprintf(stderr, "%s: Not a SimAlpha file.\n", 
	    prog_name);
    exit(1);
  }
  while(!feof(fp)){
    fgets(buf, 4096, fp);
    if(*buf=='@'){

      int ost=0; /* find the offset of dat */
      for(int i=0; i<32; i++){
	if(*(buf+i)==' '){ ost=i; break; }
      }

      ADDR_TYPE adr = get_hex(buf+1);
      DATA_TYPE dat = get_hex(buf+ost+1);
      data_t a,d; 
      a = adr;
      d = dat;
      mm->st_8byte(&a, &d, 0);
    }
  }
  fclose(fp);
}

/************************************************/
void memory_system::ld_nbyte(int n, data_t *a, 
                             data_t *d){
  if(*a%n!=0) 
    printf("*** ld_nbyte %d miss-align.\n", n);
  if(n!=1 && n!=2 && n!=4 && n!=8)
    printf("*** Error in ld_nbyte %d.\n", n);
  
  ld_8byte(a, d); /** Type Conversion **/
  if(n==8) return;

  int offset = *a & 7;
  if(n==4){
    DATA_TYPE dt = (*d>>(offset*8))&MASK32;
    *d=dt;
  }
  else if(n==2){
    DATA_TYPE dt = (*d>>(offset*8))&MASK16;
    *d=dt;
  }
  else if(n==1){
    DATA_TYPE dt = (*d>>(offset*8))&MASK08;
    *d=dt;
  }
}

/************************************************/
void memory_system::st_nbyte(int n, data_t *a, 
                             data_t *d){
  if(*a%n!=0)
    printf("*** st_nbyte %d miss-alig.\n", n);
  if(n!=1 && n!=2 && n!=4 && n!=8)
    printf("*** Error in st_nbyte %d.\n", n);

  int offset = *a & 7;
  DATA_TYPE mask=0;

  if(n==4){
    mask = ~(MASK32 << offset*8);
    DATA_TYPE dt = (*d&MASK32)<<offset*8;
    *d=dt;
  }
  else if(n==2){
    mask = ~(MASK16 << offset*8);
    DATA_TYPE dt = (*d&MASK16)<<offset*8;
    *d=dt;
  }
  else if(n==1){
    mask = ~(MASK08 << offset*8);
    DATA_TYPE dt = (*d&MASK08)<<offset*8;
    *d=dt;
  }

  st_8byte(a, d, mask); /** Type Conversion **/
}

/** Implementation of raw main memory          **/
/************************************************/
int mm_index(uint_32t adr){
  return (adr >> BLOCK_MASK_BIT);
}

/************************************************/
main_memory::main_memory(Env *ev_t){
  ev = ev_t;
  for(int i=0; i<BLOCK_TABLE_SIZE; i++){
    block_table[i]=NULL;
  }
}

/************************************************/
data_t *main_memory::allocblock(data_t *a){
  int n = BLOCK_SIZE/DATA_T_SIZE;
  data_t *ret =new data_t[n];
  for(int i=0; i<n; i++) ret[i] = 0;

  block_table[mm_index(*a)] = ret;
  if(ret==NULL){
    printf("** Error in allocblock.\n");
    exit(0);
  }
  ev->e->used_memory_block++;
  return ret;
}

/************************************************/
void main_memory::ld_8byte(data_t *a, data_t *d){
  uint_32t adr = *a;
  data_t *tt = block_table[mm_index(adr)];
  data_t *ptr = (tt!=NULL) ? tt : allocblock(a);
  int offset = (adr & BLOCK_MASK)>>DATA_T_BIT;
  *d = ptr[offset];
}

/************************************************/
void main_memory::st_8byte(data_t *a, data_t *d, 
			   DATA_TYPE msk){
  uint_32t adr = *a;
  data_t *tt = block_table[mm_index(adr)];
  data_t *ptr = (tt!=NULL) ? tt : allocblock(a);
  int offset = (adr & BLOCK_MASK)>>DATA_T_BIT;
  ptr[offset] = (msk==0) ? *d :
    (ptr[offset] & msk) | *d;
}
/************************************************/
